/*
 * jafet - printk.c
 * Copyright (C) 2014  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <linkage.h>
#include <ptrace.h>
#include <irq.h>
#include <irqflags.h>
#include <errno.h>
#include <asm_io.h>
#include <spinlock.h>
#include <special_insns.h>

#define NR_PRIORITY_IRQS	16
#define NR_IRQS			272

#define CORE_IRQ_BASE		256

#define NR_COMBINERS	4
#define NR_MUX_OUTPUTS  12

#define IRQ_UNMAPPED 0xffff

struct irqdesc {
	irq_handler_t handler;
	void *dev_id;
	const char *name;
};

struct megamod_regs {
	u32	evtflag[8];
	u32	evtset[8];
	u32	evtclr[8];
	u32	reserved0[8];
	u32	evtmask[8];
	u32	mevtflag[8];
	u32	expmask[8];
	u32	mexpflag[8];
	u32	intmux_unused;
	u32	intmux[7];
	u32	reserved1[8];
	u32	aegmux[2];
	u32	reserved2[14];
	u32	intxstat;
	u32	intxclr;
	u32	intdmask;
	u32	reserved3[13];
	u32	evtasrt;
};

struct megamod_pic {
	struct irq_domain *irqhost;
	struct megamod_regs __iomem *regs;
	raw_spinlock_t lock;
};

static struct irqdesc irqs[NR_IRQS];
static DEFINE_SPINLOCK(irq_lock);
static DEFINE_RAW_SPINLOCK(core_irq_lock);
static struct megamod_regs __iomem *pic_regs;

static struct megamod_pic mm_pic;

static void mask_core_irq(unsigned int prio)
{
	raw_spin_lock(&core_irq_lock);
	and_creg(IER, ~(1 << prio));
	raw_spin_unlock(&core_irq_lock);
}

static void unmask_core_irq(unsigned int prio)
{
	raw_spin_lock(&core_irq_lock);
	or_creg(IER, 1 << prio);
	raw_spin_unlock(&core_irq_lock);
}

static void pic_mask(unsigned int src)
{
	struct megamod_pic *pic = &mm_pic;
	u32 __iomem *evtmask = &pic->regs->evtmask[src / 32];

	raw_spin_lock(&pic->lock);
	soc_writel(soc_readl(evtmask) | (1 << (src & 31)), evtmask);
	raw_spin_unlock(&pic->lock);
}

static void pic_unmask(unsigned int src)
{
	struct megamod_pic *pic = &mm_pic;
	u32 __iomem *evtmask = &pic->regs->evtmask[src / 32];

	raw_spin_lock(&pic->lock);
	soc_writel(soc_readl(evtmask) & ~(1 << (src & 31)), evtmask);
	raw_spin_unlock(&pic->lock);
}

static irqreturn_t pic_irq_handler(int irq, void *dev_id)
{
	struct megamod_pic *pic = &mm_pic;
	unsigned int idx = irq - CORE_IRQ_BASE - 4;
	struct irqdesc *desc;
	int n;
	u32 events;

	while ((events = soc_readl(&pic->regs->mevtflag[idx])) != 0) {
		n = __ffs(events);
		irq = idx * 32 + n;

		soc_writel(1 << n, &pic->regs->evtclr[idx]);

		desc = &irqs[irq];
		if (desc->handler)
			desc->handler(irq, desc->dev_id);
	}
	return IRQ_HANDLED;
}

asmlinkage void c6x_do_IRQ(unsigned int prio, struct pt_regs *regs)
{
	struct irqdesc *desc;
	int irq;

	if (prio >= NR_PRIORITY_IRQS) {
		pr_err("invalid IRQ %d\n", prio);
		return;
	}
	irq = CORE_IRQ_BASE + prio;
	desc = &irqs[irq];
	if (desc->handler)
		desc->handler(irq, desc->dev_id);
}

/**
 * request_irq - allocate an interrupt line
 * @irq: Interrupt line to allocate
 * @handler: Function to be called when the IRQ occurs.
 * @flags: Interrupt type flags
 * @name: An ascii name for the claiming device
 * @dev_id: A cookie passed back to the handler function
 *
 * This call allocates interrupt resources and enables the
 * interrupt line and IRQ handling. From the point this
 * call is made your handler function may be invoked. Since
 * your handler function must clear any interrupt the board
 * raises, you must take care both to initialise your hardware
 * and to set up the interrupt handler in the right order.
 *
 * Dev_id must be globally unique. Normally the address of the
 * device data structure is used as the cookie. Since the handler
 * receives this value it makes sense to use it.
 *
 * If your interrupt is shared you must pass a non NULL dev_id
 * as this is required when freeing the interrupt.
 *
 * Flags:
 *
 * IRQF_SHARED		Interrupt is shared
 */
int request_irq(unsigned int irq, irq_handler_t handler,
		unsigned long flags, const char *name, void *dev_id)
{
	struct irqdesc *desc = irqs + irq;
	int ret;

	if (irq >= NR_IRQS)
		return -EINVAL;

	spin_lock(&irq_lock);
	ret = -EBUSY;
	if (desc->handler) {
		printk("irq %d handler: %p\n", irq, desc->handler);
		goto out_unlock;
	}
	desc->name = name;
	desc->dev_id = dev_id;
	desc->handler = handler;

	if (irq < 256)
		pic_unmask(irq);
	ret = 0;

out_unlock:
	spin_unlock(&irq_lock);
	return ret;
}

static void set_megamod_mux(struct megamod_pic *pic, int src, int output)
{
	int index, offset;
	u32 val;

	if (src < 0 || src >= (NR_COMBINERS * 32))
		return;

	/* four mappings per mux register */
	index = output / 4;
	offset = (output & 3) * 8;

	val = soc_readl(&pic->regs->intmux[index]);
	val &= ~(0xff << offset);
	val |= src << offset;
	soc_writel(val, &pic->regs->intmux[index]);
}

void irq_init(void)
{
	struct megamod_pic *pic = &mm_pic;
	int mapping[NR_MUX_OUTPUTS];
	unsigned int i;

	/* Mask all priority IRQs */
	and_creg(IER, ~0xfff0);
	i = get_creg(IER);

	pic->regs = (void *) 0x01800000;

	for (i = 0; i < NR_IRQS; i++) {
		struct irqdesc *desc = &irqs[i];

		desc->handler = NULL;
	}


	/* Initialize MUX map */
	for (i = 0; i < ARRAY_SIZE(mapping); i++)
		mapping[i] = IRQ_UNMAPPED;

	for (i = 0; i < NR_COMBINERS; i++) {
		unsigned int hwirq;

		hwirq = 4 + i;
		mapping[i] = i;

		/* mask and clear all events in combiner */
		soc_writel(~0, &pic->regs->evtmask[i]);
		soc_writel(~0, &pic->regs->evtclr[i]);
	}

	/* Finally, set up the MUX registers */
	for (i = 0; i < NR_MUX_OUTPUTS; i++) {
		if (mapping[i] != IRQ_UNMAPPED) {
			pr_debug("setting mux %d to priority %d\n",
				 mapping[i], i + 4);
			set_megamod_mux(pic, mapping[i], i);
		}
	}

	for (i = 0; i < NR_COMBINERS; i++) {
		request_irq(CORE_IRQ_BASE + 4 + i, pic_irq_handler,
				0, "pic-irq", NULL);
		unmask_core_irq(4 + i);
	}

	/* Clear all general IRQ flags */
	set_creg(ICR, 0xfff0);

	arch_local_irq_enable();
}
