/*
 * noisecom - kmalloc.c
 *
 * dynamic memory allocator
 * 
 * Copyright (C) 2011, 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>

#define KMALLOC_ALIGNMENT	8
#define KMALLOC_MASK		(KMALLOC_ALIGNMENT - 1)

/*
 * currently for all allocation L2 RAM is used. This memory can be used
 * for all purposes, including DMA transfers.
 *
 * FIXME: check this region,
 */
static unsigned int l1_brk = 0x11f00000;
static unsigned int l1_max = 0x11f04000;
extern void *kmalloc_brk;
extern void *kmalloc_brk_limit;
static unsigned int l2_brk = (unsigned int) &kmalloc_brk;
static unsigned int l2_max = (unsigned int) &kmalloc_brk_limit;

void *kmalloc_l1_floatv(unsigned int n)
{
	void *p;
	size_t size;

	size = n * sizeof(float);
	if (size >= 65536)
		return NULL;

	if (l1_brk + size > l1_max) {
		printk("error: Cache L1 allocation failure\n");
		return NULL;
	}
	
	p = (void *) l1_brk;
	return p;
}

void *kmalloc_l1(size_t size, gfp_t flags)
{
	void *p;

	if (size >= 65536)
		return NULL;

	if (l1_brk + size > l1_max) {
		printk("error: Cache L1 allocation failure\n");
		return NULL;
	}
	
	p = (void *) l1_brk;
	/*
	 * Cache L1 coloring
	 *
	 * The Cache L1 in C6x brain-damaged. Like in Pentium
	 * it's banked and only one value can be read from each
	 * bank, there are 8 32-bit banks. The major difference
	 * is that Pentium executed up to 2 instructions per clock
	 * and both could be loads, the C6x may execute up to
	 * 8 instructions per clock with only 2 loads and in
	 * case of Cache L1 bank conflict only one. Because
	 * the DSP code uses time-regresive vectors we cannot
	 * avoid conflicts! We still do some coloring in hope
	 * that we will avoid bank conflicts in all vector
	 * operations doubling execution time.
	 */
	if (flags & __GFP_COLOR4)
		l1_brk += (size + 4) & ~3;
	else
		l1_brk += (size + KMALLOC_ALIGNMENT - 1) & ~KMALLOC_MASK;
	return p;
}

void *kmalloc_l2(size_t size)
{
	void *p;

	if (size >= 65536)
		return NULL;

	if (l2_brk + size > l2_max) {
		printk("error: Cache L2 allocation failure\n");
		return NULL;
	}
	
	p = (void *) l2_brk;
	l2_brk += (size + KMALLOC_ALIGNMENT - 1) & ~KMALLOC_MASK;
	return p;
}

void *kmalloc(size_t size, gfp_t flags)
{
	if (flags & __GFP_L1)
		kmalloc_l1(size, flags);
	return kmalloc_l2(size);
}

void kfree(const void *p)
{
}
