/*
 * np-dsp - mcasp.c
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#define pr_fmt(fmt) "mcasp: " fmt
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <errno.h>
#include <edma.h>
#include <asm_io.h>
#include <timex.h>

#if 1
#define CONFIG_DA830 	1
#endif

#define MCASP_CHANNELS		2

struct mcasp_params {
	unsigned int channels;
	unsigned int downsample;
	unsigned int ser_rx_mask;
	unsigned int ser_tx_mask;
	unsigned int num_serializers;
};

struct mcasp_resources {
	unsigned long regs_addr;
	unsigned long data_addr;
	unsigned int rx_channel;
	unsigned int tx_channel;
};

struct mcasp {
	u32 *rx_buf;
	u32 *tx_buf;
	u8 __iomem *regs;
	unsigned int rx_dma;
	unsigned int tx_dma;
	const struct mcasp_params *params;

	unsigned long seq;
	float rx[MCASP_CHANNELS];
	float tx[MCASP_CHANNELS];
};
static struct mcasp *mcasp;

enum {
	MCASP_REVID		= 0x0000,
	MCASP_PWRDEMU		= 0x0004,
	MCASP_PFUNC		= 0x0010,
	MCASP_PDIR		= 0x0014,
	MCASP_PDOUT		= 0x0018,
	MCASP_PDIN		= 0x001C,
	MCASP_PDCLR		= 0x0020,
	MCASP_GBLCTL		= 0x0044,
	MCASP_AMUTE		= 0x0048,
	MCASP_DLBCTL		= 0x004C,
	MCASP_DITCTL		= 0x0050,
	MCASP_RGBLCTL		= 0x0060,
	MCASP_RMASK		= 0x0064,
	MCASP_RFMT		= 0x0068,
	MCASP_AFSRCTL		= 0x006C,
	MCASP_ACLKRCTL		= 0x0070,
	MCASP_AHCLKRCTL		= 0x0074,
	MCASP_RTDM		= 0x0078,
	MCASP_RINTCTL		= 0x007C,
	MCASP_RSTAT		= 0x0080,
	MCASP_RSLOT		= 0x0084,
	MCASP_RCLKCHK		= 0x0088,
	MCASP_REVTCTL		= 0x008C,
	MCASP_XGBLCTL		= 0x00A0,
	MCASP_XMASK		= 0x00A4,
	MCASP_XFMT		= 0x00A8,
	MCASP_AFSXCTL		= 0x00AC,
	MCASP_ACLKXCTL		= 0x00B0,
	MCASP_AHCLKXCTL		= 0x00B4,
	MCASP_XTDM		= 0x00B8,
	MCASP_XINTCTL		= 0x00BC,
	MCASP_XSTAT		= 0x00C0,
	MCASP_XSLOT		= 0x00C4,
	MCASP_XCLKCHK		= 0x00C8,
	MCASP_XEVTCTL		= 0x00CC,
	MCASP_DITCSRA0		= 0x0100,
	MCASP_DITCSRA1		= 0x0104,
	MCASP_DITCSRA2		= 0x0108,
	MCASP_DITCSRA3		= 0x010C,
	MCASP_DITCSRA4		= 0x0110,
	MCASP_DITCSRA5		= 0x0114,
	MCASP_DITCSRB0		= 0x0118,
	MCASP_DITCSRB1		= 0x011C,
	MCASP_DITCSRB2		= 0x0120,
	MCASP_DITCSRB3		= 0x0124,
	MCASP_DITCSRB4		= 0x0128,
	MCASP_DITCSRB5		= 0x012C,
	MCASP_DITUDRA0		= 0x0130,
	MCASP_DITUDRA1		= 0x0134,
	MCASP_DITUDRA2		= 0x0138,
	MCASP_DITUDRA3		= 0x013C,
	MCASP_DITUDRA4		= 0x0140,
	MCASP_DITUDRA5		= 0x0144,
	MCASP_DITUDRB0		= 0x0148,
	MCASP_DITUDRB1		= 0x014C,
	MCASP_DITUDRB2		= 0x0150,
	MCASP_DITUDRB3		= 0x0154,
	MCASP_DITUDRB4		= 0x0158,
	MCASP_DITUDRB5		= 0x015C,
	MCASP_SRCTL0		= 0x0180,
	MCASP_SRCTL1		= 0x0184,
	MCASP_SRCTL2		= 0x0188,
	MCASP_SRCTL3		= 0x018C,
	MCASP_SRCTL4		= 0x0190,
	MCASP_SRCTL5		= 0x0194,
	MCASP_SRCTL6		= 0x0198,
	MCASP_SRCTL7		= 0x019C,
	MCASP_SRCTL8		= 0x01A0,
	MCASP_SRCTL9		= 0x01A4,
	MCASP_SRCTL10		= 0x01A8,
	MCASP_SRCTL11		= 0x01AC,
	MCASP_SRCTL12		= 0x01B0,
	MCASP_SRCTL13		= 0x01B4,
	MCASP_SRCTL14		= 0x01B8,
	MCASP_SRCTL15		= 0x01BC,
	MCASP_XBUF0		= 0x0200,
	MCASP_XBUF1		= 0x0204,
	MCASP_XBUF2		= 0x0208,
	MCASP_XBUF3		= 0x020C,
	MCASP_XBUF4		= 0x0210,
	MCASP_XBUF5		= 0x0214,
	MCASP_XBUF6		= 0x0218,
	MCASP_XBUF7		= 0x021C,
	MCASP_XBUF8		= 0x0220,
	MCASP_XBUF9		= 0x0224,
	MCASP_XBUF10		= 0x0228,
	MCASP_XBUF11		= 0x022C,
	MCASP_XBUF12		= 0x0230,
	MCASP_XBUF13		= 0x0234,
	MCASP_XBUF14		= 0x0238,
	MCASP_XBUF15		= 0x023C,
	MCASP_RBUF0		= 0x0280,
	MCASP_RBUF1		= 0x0284,
	MCASP_RBUF2		= 0x0288,
	MCASP_RBUF3		= 0x028C,
	MCASP_RBUF4		= 0x0290,
	MCASP_RBUF5		= 0x0294,
	MCASP_RBUF6		= 0x0298,
	MCASP_RBUF7		= 0x029C,
	MCASP_RBUF8		= 0x02A0,
	MCASP_RBUF9		= 0x02A4,
	MCASP_RBUF10		= 0x02A8,
	MCASP_RBUF11		= 0x02AC,
	MCASP_RBUF12		= 0x02B0,
	MCASP_RBUF13		= 0x02B4,
	MCASP_RBUF14		= 0x02B8,
	MCASP_RBUF15		= 0x02BC,
};

#define MCASP_XRSRCTL_BASE_REG	0x180
#define MCASP_XRSRCTL_REG(n)	(MCASP_XRSRCTL_BASE_REG + (n << 2))

#define MCASP_GBLCTL_XFRST        0x00001000
#define MCASP_GBLCTL_XSMRST       0x00000800
#define MCASP_GBLCTL_XSRCLR       0x00000400
#define MCASP_GBLCTL_XHCLKRST     0x00000200
#define MCASP_GBLCTL_XCLKRST      0x00000100
#define MCASP_GBLCTL_RFRST        0x00000010
#define MCASP_GBLCTL_RSMRST       0x00000008
#define MCASP_GBLCTL_RSRCLR       0x00000004
#define MCASP_GBLCTL_RHCLKRST     0x00000002
#define MCASP_GBLCTL_RCLKRST      0x00000001

static const struct mcasp_resources mcasp_resources[] = {
	{ 0x01d00000, 0x01d02000, 0, 1 },
	{ 0x01d04000, 0x01d06000, 2, 3 },
};

static const struct mcasp_params mcasp_params = {
	.channels		= 2,
	.downsample		= 24,
	.ser_rx_mask		= 0x4000,
	.ser_tx_mask		= 0x8000,
	.num_serializers	= 16,
};

static const struct mcasp_params mcasp_params_da830 = {
	.channels		= 2,
	.downsample		= 24,
	.ser_rx_mask		= 0x0001,
	.ser_tx_mask		= 0x0020,
	.num_serializers	= 12,
};

/**
 * float_to_raw - convert floating point audio data to raw format
 * @x: sample to convert, normalized to [-1; 1) range
 */
static s32 float_to_raw(float x)
{
	s32 y = ((s32)(65536.0f * 16384.0f * x));

	if (y > 0x3fffffff)
		y = 0x3fffffff;
	else if (y < -0x40000000)
		y = -0x40000000;
	return y << 1;
}

/**
 * raw_to_float - convert raw audio data to floating point format
 * @x: sample to convert
 *
 * This function returns sample value in floating point format normalized
 * to [-1; 1) range.
 */
static float raw_to_float(s32 x)
{
	return ((float) (s32) x) * 1.0f / (65536.0f * 32768.0f);
}

/**
 * float_to_s16 - convert floating point audio data to raw format
 * @x: sample to convert, normalized to [-1; 1) range
 */
static s16 float_to_s16(float x)
{
	return (s16)(32767.0f * x);
}

static float ari2_filter(float *y, const float *w, float x)
{
	float t;

	t = y[0] + w[0] * (y[0] - x) + w[1] * (y[1] - x);
	y[1] = y[0];
	y[0] = t;
	return t;
}

static float ari4_filter(float *y, const float *w, float x)
{
	float t;

	t = ari2_filter(y, w, x);
	t = ari2_filter(y + 2, w + 2, t);
	return t;
}

/*
 * Antialiasing filter
 *
 * Passband: 853 Hz	(1138 Hz for 10 dB)
 * Attenuation: 40 dB
 * Max delay: 0.521 ms
 */
static const float antialias_w[] = {
	0.94977f, -0.95355f,
	0.88776f, -0.89142f
};

static float hari2_filter(float *y, const float *w, float x)
{
	float y0, y1, y2;

	y2 = y[1];
	y1 = y[0];
	y0 = y1 + w[0] * (y1 - x) + w[1] * (y2 - x);
	y[1] = y1;
	y[0] = y0;
	return w[2] * (y0 - 2 * y1 + y2);
}

/*
 * Highpass filter
 *
 * Passband: 70 Hz	(40 Hz for 10 dB)
 */
static const float highpass_w[] = {
	0.84478f, -0.85599f, 82.5f
};

static unsigned long tx_maxt;

extern int pu_shm_store(const void *buffer, size_t len);

static void mcasp_dma_tx_callback(unsigned channel, u16 ch_status, void *data)
{
	static float y_out[2][4];
	static float y_in[2][4];
	static float in_offset_state[2][2];
	s16 shm_data[4];
	struct mcasp *p = data;
	unsigned int downsample = p->params->downsample;
	unsigned int channels = p->params->channels;
	unsigned int i;
	unsigned long seq;
	unsigned int offset;
	unsigned long t;
	float tx0;
	float tx1;
	float rx0 = 0;
	float rx1 = 0;

	mb();
	t = get_cycles();
	offset = 0;
	tx0 = p->tx[0];
	tx1 = p->tx[1];
	for (i = 0; i < downsample; i++) {
		float x;

		x = ari4_filter(y_out[0], antialias_w, tx0);
		p->tx_buf[offset + 0] = float_to_raw(x);

		x = ari4_filter(y_out[1], antialias_w, tx1);
		p->tx_buf[offset + 1] = float_to_raw(x);

		x = raw_to_float(p->rx_buf[offset + 0]);
		rx0 = ari4_filter(y_in[0], antialias_w, x);

		x = raw_to_float(p->rx_buf[offset + 1]);
		rx1 = ari4_filter(y_in[1], antialias_w, x);
		offset += channels;
	}

	rx0 = hari2_filter(in_offset_state[0], highpass_w, rx0);
	rx1 = hari2_filter(in_offset_state[1], highpass_w, rx1);

	p->rx[0] = rx0;
	p->rx[1] = rx1;

	shm_data[0] = float_to_s16(rx0);
	shm_data[1] = float_to_s16(rx1);
	shm_data[2] = float_to_s16(tx0);
	shm_data[3] = float_to_s16(tx1);

	pu_shm_store(shm_data, sizeof(shm_data));

	seq = p->seq;
	seq++;

	smp_wmb();
	ACCESS_ONCE(p->seq) = seq;
	t = get_cycles() - t;

	if (t > tx_maxt) {
		tx_maxt = t;
		pr_info("mcasp time: %ld\n", t);
	}
}

static void mcasp_set_and_wait(struct mcasp *p, u32 mask,
		unsigned int reg)
{
	u32 x;

	x = soc_readl(p->regs + reg);
	x |= mask;
	soc_writel(x, p->regs + reg);

	do {
		x = soc_readl(p->regs + reg);
	} while (!(x & mask));
}

struct mcasp_frame {
	u32 seq;
	float data[];
};

ssize_t mcasp_read(void *buf, size_t count)
{
	struct mcasp *p = mcasp;
	struct mcasp_frame *frame = buf;
	unsigned int n;
	static unsigned long i;
	static unsigned long old_seq;
	dma_addr_t src, dst;
	unsigned long seq;

	if (count < sizeof(*frame))
		return -EINVAL;
	n = (count - sizeof(*frame)) / sizeof(*frame->data);
	if (n > MCASP_CHANNELS)
		n = MCASP_CHANNELS;

	i++;
	seq = ACCESS_ONCE(p->seq);
	if (i >= 1000000) {
		mb();
		edma_get_position(64, &src, &dst);
		pr_info("timeout: seq: %ld  src: %08lx, dst: %08lx\n",
				seq, src, dst);
		edma_get_position(65, &src, &dst);
		pr_info("timeout: src: %08lx, dst: %08lx\n",
				src, dst);
		pr_info("%08x %08x %08x %08x\n",
				p->rx_buf[0], p->rx_buf[1],
				p->rx_buf[2], p->rx_buf[3]);

		pr_info("timeout: RSTAT: %08x  XSTAT: %08x\n",
				soc_readl(p->regs + MCASP_RSTAT),
				soc_readl(p->regs + MCASP_XSTAT));
		
		pr_info("timeout: RCLKCHK: %08x  XCLKCHK, %08x\n",
				soc_readl(p->regs + MCASP_RCLKCHK),
				soc_readl(p->regs + MCASP_XCLKCHK));
	
		soc_writel(0x0000ffff, p->regs + MCASP_XSTAT);
		soc_writel(0x0000ffff, p->regs + MCASP_RSTAT);
		i = 0;
	}
	if (seq == old_seq)
		return -EAGAIN;


	smp_rmb();
	old_seq = seq;
	i = 0;

	frame->seq = seq;
	for (i = 0; i < n; i++)
		frame->data[i] = p->rx[i];

	seq = ACCESS_ONCE(p->seq);
	if (seq != old_seq)
		pr_warn("drop\n");

	return sizeof(*frame) + n * sizeof(*frame->data);
}

ssize_t mcasp_write(const void *buf, size_t count)
{
	const struct mcasp_frame *frame = buf;
	struct mcasp *p = mcasp;
	unsigned int n;
	unsigned int i;

	n = (count - sizeof(*frame)) / sizeof(*frame->data);
	if (n > MCASP_CHANNELS)
		n = MCASP_CHANNELS;
	for (i = 0; i < n; i++)
		p->tx[i] = frame->data[i];
	return sizeof(*frame) + n * sizeof(*frame->data);
}

static int mcasp_init(struct mcasp *p, const struct mcasp_resources *res,
		const struct mcasp_params *params)
{
	struct edmacc_param rx_params;
	struct edmacc_param tx_params;
	unsigned int frame_size = sizeof(*p->rx_buf);
	unsigned int buf_size;
	unsigned int slot;
	unsigned int ch;
	unsigned int i;
	int ret;
	u32 rev;

	p->params = params;

	p->regs = __io_virt(res->regs_addr);

	rev = soc_readl(p->regs + MCASP_REVID);
	pr_info("revision 0x%08x at 0x%08x\n", rev, res->regs_addr);

	pr_info("channels: %d, downsample: %d\n", params->channels,
			params->downsample);

	buf_size = frame_size * params->downsample;
	buf_size *= params->channels;

	pr_info("frame size: %d, buffer size: %d\n", frame_size, buf_size);

	soc_writel(0, p->regs + MCASP_GBLCTL);
	soc_writel(0, p->regs + MCASP_RGBLCTL);
	soc_writel(0, p->regs + MCASP_XGBLCTL);

	p->rx_buf = kmalloc(buf_size, GFP_KERNEL | __GFP_DMA);
	p->tx_buf = kmalloc(buf_size, GFP_KERNEL | __GFP_DMA);
	ret = -ENOMEM;
	if (!p->rx_buf || !p->tx_buf)
		goto out_free;

	pr_info("buffers: rx: 0x%08x, tx: 0x%08x\n", p->rx_buf, p->tx_buf);

	memset(p->rx_buf, 0xaa, buf_size);
	memset(p->tx_buf, 0, buf_size);
	mb();

	ch = edma_alloc_channel(res->rx_channel, NULL, NULL,
			EVENTQ_DEFAULT);
	p->rx_dma = ch;

	edma_set_default_opts(ch);
	edma_set_src(ch, virt_to_dma(res->data_addr), INCR, W8BIT);
	edma_set_dest(ch, virt_to_dma(p->rx_buf), INCR, W8BIT);
	edma_set_src_index(ch, 0, 0);
	edma_set_dest_index(ch, frame_size, 0);

	edma_set_transfer_params(ch, frame_size,
			params->channels * params->downsample, 1, 0, ASYNC);
	slot = edma_alloc_slot(0, 64);
	if (slot != 64) {
		printk("slot error %d\n", slot);
	}

	edma_read_slot(ch, &rx_params);
	rx_params.opt |= EDMA_TCC(res->rx_channel);
	rx_params.link_bcntrld &= ~0xffff;
	rx_params.link_bcntrld |= slot << 5;
	edma_write_slot(ch, &rx_params);
	edma_write_slot(slot, &rx_params);

	edma_start(ch);
	ch = edma_alloc_channel(res->tx_channel, mcasp_dma_tx_callback,
			p, EVENTQ_DEFAULT);
	p->tx_dma = ch;

	edma_set_default_opts(ch);
	edma_set_dest(ch, virt_to_dma(res->data_addr), INCR, W8BIT);
	edma_set_src(ch, virt_to_dma(p->tx_buf), INCR, W8BIT);
	edma_set_src_index(ch, frame_size, 0);
	edma_set_dest_index(ch, 0, 0);

	edma_set_transfer_params(ch, frame_size,
			params->channels * params->downsample, 1, 0, ASYNC);

	slot = edma_alloc_slot(0, 65);
	if (slot != 65) {
		printk("slot error %d\n", slot);
	}

	edma_read_slot(ch, &tx_params);
	tx_params.opt |= EDMA_TCC(res->tx_channel);
	tx_params.link_bcntrld &= ~0xffff;
	tx_params.link_bcntrld |= slot << 5;
	edma_write_slot(ch, &tx_params);
	edma_write_slot(slot, &tx_params);
	edma_start(ch);

	soc_writel(0, p->regs + MCASP_GBLCTL);
	soc_writel(0, p->regs + MCASP_RGBLCTL);
	soc_writel(0, p->regs + MCASP_XGBLCTL);

	/* free running */
	soc_writel(1, p->regs + MCASP_PWRDEMU);

	soc_writel(0xffffffff, p->regs + MCASP_RMASK);
	soc_writel(0x000180f0, p->regs + MCASP_RFMT);
	soc_writel(0x00000110, p->regs + MCASP_AFSRCTL);
	soc_writel(0x0000008f, p->regs + MCASP_ACLKRCTL);
	soc_writel(0x00000000, p->regs + MCASP_AHCLKRCTL);
	soc_writel(0x00000003, p->regs + MCASP_RTDM);
	soc_writel(0x00000020, p->regs + MCASP_RINTCTL);
	soc_writel(0x00ff0008, p->regs + MCASP_RCLKCHK);

	soc_writel(0xffffffff, p->regs + MCASP_XMASK);
	soc_writel(0x000180f0, p->regs + MCASP_XFMT);
	soc_writel(0x00000110, p->regs + MCASP_AFSXCTL);
	soc_writel(0x0000008f, p->regs + MCASP_ACLKXCTL);
	soc_writel(0x00000000, p->regs + MCASP_AHCLKXCTL);
	soc_writel(0x00000003, p->regs + MCASP_XTDM);
	soc_writel(0x00000020, p->regs + MCASP_XINTCTL);
	soc_writel(0x00ff0008, p->regs + MCASP_XCLKCHK);

	for (i = 0; i < params->num_serializers; i++) {
		if (params->ser_tx_mask & BIT(i))
			soc_writel(0xd, p->regs + MCASP_XRSRCTL_REG(i));
		else if (params->ser_rx_mask & BIT(i))
			soc_writel(0xe, p->regs + MCASP_XRSRCTL_REG(i));
		else
			soc_writel(0x0, p->regs + MCASP_XRSRCTL_REG(i));
	}

	soc_writel(0, p->regs + MCASP_PFUNC);
	soc_writel(params->ser_tx_mask, p->regs + MCASP_PDIR);
	soc_writel(0, p->regs + MCASP_DITCTL);
	soc_writel(0, p->regs + MCASP_DLBCTL);
	soc_writel(0, p->regs + MCASP_AMUTE);

	mcasp_set_and_wait(p, MCASP_GBLCTL_XHCLKRST, MCASP_XGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_RHCLKRST, MCASP_RGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_XCLKRST, MCASP_XGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_RCLKRST, MCASP_RGBLCTL);

	soc_writel(0x0000ffff, p->regs + MCASP_XSTAT);
	soc_writel(0x0000ffff, p->regs + MCASP_RSTAT);

	mcasp_set_and_wait(p, MCASP_GBLCTL_XSRCLR, MCASP_XGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_RSRCLR, MCASP_RGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_XSMRST, MCASP_XGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_RSMRST, MCASP_RGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_XFRST, MCASP_XGBLCTL);
	mcasp_set_and_wait(p, MCASP_GBLCTL_RFRST, MCASP_RGBLCTL);

	pr_info("MCASP is running\n");
	return 0;

out_free:
	kfree(p->rx_buf);
	kfree(p->tx_buf);
	return ret;
}

void mcasp_module_init(void)
{
	struct mcasp *p;

	p = kmalloc(sizeof(*p), GFP_KERNEL | __GFP_L1);
	memset(p, 0, sizeof(*p));
	mcasp = p;
#if CONFIG_DA830
	mcasp_init(p, &mcasp_resources[1], &mcasp_params_da830);
#else
	mcasp_init(p, &mcasp_resources[0], &mcasp_params);
#endif
}
