/*
 *  Port on Texas Instruments TMS320C6x architecture
 *
 *  Copyright (C) 2004, 2006, 2009, 2010, 2011 Texas Instruments Incorporated
 *  Author: Aurelien Jacquiot (aurelien.jacquiot@jaluna.com)
 *  
 *  Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <errno.h>
#include <linkage.h>
#include <ptrace.h>
#include <mcasp.h>

#define __NR_READ	0
#define __NR_WRITE	1
#define __NR_IOCTL	16

struct file;

struct file_operations {
	ssize_t (*read)(struct file *, char __user *, size_t, loff_t *);
	ssize_t (*write)(struct file *, const char __user *, size_t, loff_t *);
	long (*unlocked_ioctl)(struct file *, unsigned int, unsigned long);
};

struct file {
	struct file_operations *f_op;
	loff_t f_pos;
};

static ssize_t log_write(struct file *file, const char __user *buf,
		size_t count, loff_t *off)
{
	static char line[129];

	if (count > sizeof(line))
		count = sizeof(line);
	memcpy(line, buf, count);
	line[count] = '\0';
	printk("%s", line);
	return count;
}

static struct file_operations log_file_operations = {
	.write		= log_write,
};

static struct file log_file = {
	.f_op		= &log_file_operations,
};

static ssize_t mcasp_dev_read(struct file *file, char __user *buf,
		size_t count, loff_t *off)
{
	return mcasp_read(buf, count);
}

static ssize_t mcasp_dev_write(struct file *file, const char __user *buf,
		size_t count, loff_t *off)
{
	return mcasp_write(buf, count);
}

static struct file_operations mcasp_file_operations = {
	.read		= mcasp_dev_read,
	.write		= mcasp_dev_write,
};

static struct file mcasp_file = {
	.f_op		= &mcasp_file_operations,
};

static struct file *file_table[] = {
	&mcasp_file,
	&mcasp_file,
	&log_file,
};

static struct file *fget(unsigned int fd)
{
	struct file *file = NULL;

	if (fd < ARRAY_SIZE(file_table))
		file = file_table[fd];
	return file;
}

static void fput(struct file *file)
{
}

static int sys_read(unsigned int fd, char *buf, size_t count)
{
	struct file *file;
	int ret;

	ret = -EBADF;
	file = fget(fd);
	if (file) {
		loff_t pos = file->f_pos;

		ret = -EINVAL;
		if (file->f_op && file->f_op->read) {
			ret = file->f_op->read(file, buf, count, &pos);
			if (ret >= 0)
				file->f_pos = pos;
		}
		fput(file);
	}
	return ret;
}

static int sys_write(unsigned int fd, const char *buf, unsigned long count)
{
	struct file *file;
	int ret;

	ret = -EBADF;
	file = fget(fd);
	if (file) {
		loff_t pos = file->f_pos;

		ret = -EINVAL;
		if (file->f_op && file->f_op->write) {
			ret = file->f_op->write(file, buf, count, &pos);
			if (ret >= 0)
				file->f_pos = pos;
		}
		fput(file);
	}
	return ret;
}

static int sys_ioctl(int fd, unsigned long request, unsigned long arg)
{
	struct file *file;
	int ret;

	ret = -EBADF;
	file = fget(fd);
	if (file) {
		ret = -EINVAL;
		if (file->f_op && file->f_op->unlocked_ioctl)
			ret = file->f_op->unlocked_ioctl(file, request, arg);
		fput(file);
	}
	return ret;
}

asmlinkage void system_call(struct pt_regs *regs)
{
	int ret = -ENOSYS;
	unsigned int syscall = regs->b0;

	switch (syscall) {
	case __NR_READ:
		ret = sys_read(regs->a4, (char *) regs->b4, regs->a6);
		break;
	case __NR_WRITE:
		ret = sys_write(regs->a4, (const char *) regs->b4, regs->a6);
		break;
	case __NR_IOCTL:
		ret = sys_ioctl(regs->a4, regs->b4, regs->a6);
		break;
	}

	regs->a4 = ret;
}
